/*
 * Copyright (c) 2009-2015 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 ** ivmobMotionTaskCodels.c
 **
 ** Codels called by execution task ivmobMotionTask
 **
 ** Author: Matthieu Herrb
 ** Date: April 2009 for rmp400, updated for ivmob, Dec 2013
 **
 **/
#include <sys/types.h>
#include <errno.h>
#include <math.h>
#include <pthread.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>

#include <portLib.h>
#include <h2mathLib.h>
#include "h2timeLib.h"

#include <genPos/genPosStruct.h>
#include <contlaw/contLaw.h>
#include <ivmob/ivmob.h>

#include "ivmobConst.h"
#include "server/ivmobHeader.h"

#include "odoProba.h"

#ifdef GYRO
#include <gyroLib/gyro.h>
#endif

static POSTER_ID pomPosterId = NULL;
struct ivmob *ivmob = NULL;
#ifdef GYRO
static struct GYRO_DATA *gyroId = NULL;
#endif

extern STATUS ivmobOdo3d(int *report, double period);

static void ivmobDataUpdate(IVMOB_STATUS_STR *);
static STATUS ivmobVelocitySet(double, double, int *);
static STATUS initPomPosters(void);
static STATUS updatePomPosters(void);
static STATUS ivmobVelocityGet(void);

/*------------------------------------------------------------------------
 *
 * initOdoAndAsserv  --  Initialization codel (fIDS, ...)
 *
 * Description:
 *
 * Returns:    OK or ERROR
 */

STATUS
initOdoAndAsserv(int *report)
{
	GENPOS_CART_STATE *robot;
	GENPOS_CART_REF *refId;
	GENPOS_CART_CMD_PARAM *cmdId;
	IVMOB_KINEMATICS_STR *kinematics;
	IVMOB_GYRO *gyro;
	IVMOB_GYRO_ASSERV *gyroAsserv;

	/* Kinematics */
	kinematics = &ivmobDataStrId->kinematics;
	kinematics->wheelRadius = IVMOB_DEFAULT_WHEEL_RADIUS;
	kinematics->axisWidth = IVMOB_DEFAULT_WHEEL_TRACK_WIDTH;

	/* configuration */
	robot = &ivmobDataStrId->robot;
	robot->xRob = 0.;
	robot->yRob = 0.;
	robot->theta = 0.;
	robot->xRef = 0;
	robot->yRef = 0;
	robot->v = 0;
	robot->w = 0;

	/* Ref Data */
	refId = &ivmobDataStrId->ref;
	refId->dataType = GENPOS_POS_AND_SPEED_DATA;
	refId->backFlag = GENPOS_FORWARD_MOTION;
	refId->x = robot->xRef;
	refId->y = robot->yRef;
	refId->theta = robot->theta;
	refId->v = 0.;
	refId->w = 0.;

	refId->vmax = IVMOB_DEFAULT_MAXIMUM_VELOCITY;
	refId->wmax = IVMOB_DEFAULT_MAXIMUM_YAW_RATE;
	refId->linAccelMax = IVMOB_DEFAULT_MAXIMUM_ACCEL;
	refId->angAccelMax = IVMOB_DEFAULT_MAXIMUM_YAW_ACCEL;

	/* Control parameters */
	cmdId = &ivmobDataStrId->cmd;
	cmdId->distPoint = IVMOB_DIST_TO_CONTROLLED_POINT;
	cmdId->KpLongit = IVMOB_KP_LONGIT;
	cmdId->KiLongit = IVMOB_KI_LONGIT;
	cmdId->KpTransv = IVMOB_KP_TRANSV;
	cmdId->KiTransv = IVMOB_KI_TRANSV;

	cmdId->KpAng = IVMOB_KP_ANG;
	cmdId->KiAng = IVMOB_KI_ANG;

	cmdId->longitMaxGap = IVMOB_LONGIT_MAX_GAP;
	cmdId->transvMaxGap = IVMOB_TRANSV_MAX_GAP;
	cmdId->angMaxGap    = IVMOB_ANG_MAX_GAP;

	cmdId->longitErrorMax = IVMOB_LONGIT_ERROR_MAX;
	cmdId->transvErrorMax = IVMOB_TRANSV_ERROR_MAX;
	cmdId->angErrorMax = IVMOB_ANG_ERROR_MAX;

	cmdId->longitMaxErrSum = IVMOB_LONGIT_MAX_ERR_SUM;
	cmdId->transvMaxErrSum = IVMOB_TRANSV_MAX_ERR_SUM;
	cmdId->angMaxErrSum = IVMOB_ANG_MAX_ERR_SUM;

	ivmobDataStrId->vCommand = ivmobDataStrId->wCommand = 0.0;

	/* gyro */
	gyro = &ivmobDataStrId->gyro;
	gyro->params.mode = IVMOB_GYRO_ON;
#ifdef GYRO
	gyro->gyroOn = FALSE;
#endif
	gyro->gyroToRobotOffset = 0.0;
	gyro->gyroTheta = 0.0;

	/* gyro asserv */
	gyroAsserv = &ivmobDataStrId->gyroAsserv;
	gyroAsserv->enabled = 0;
	gyroAsserv->first = 1;
	gyroAsserv->straight = 0;

	/* max accel */
	ivmobDataStrId->maxAccel.prev_vel_command = 0.;
	ivmobDataStrId->maxAccel.prev_vel_command_t = -1.;

	/* Initialize Pom */
	initPomPosters();

	return OK;
}

/*------------------------------------------------------------------------
 *
 * endOdoAndAsserv  --  Termination codel
 *
 * Description: stop the motors and disconnect from robot
 *
 * Returns:    OK or ERROR
 */

STATUS
endOdoAndAsserv(void)
{
	if (ivmob == NULL)
		return OK;
	ivmobCmdInactive(ivmob, 0);
	ivmobRqstPeriodicMessages(ivmob, 0, 0, 0, 0, 0, 0, 0);
	ivmobEnd(ivmob);
	ivmob = NULL;
	return OK;
}

/*----------------------------------------------------------------------*/
/*
 * ivmobDataUpdate - copy data from feedback structure to the IDS
 */

static void
ivmobDataUpdate(IVMOB_STATUS_STR *status)
{
	static int prevMode = -1;
	static int prevOrigin = -1;

	ivmob_lock(ivmob);
	if (ivmob->operational_mode != prevMode) {
		printf("** mode change: %s -> %s\n",
		    ivmob_operating_mode(prevMode),
		    ivmob_operating_mode(ivmob->operational_mode));
	}
	status->operational_mode = ivmob->operational_mode;

	/* Check for manual mode */
	if (ivmob->cmd_origin != prevOrigin) {
		printf("** command origin: %s\n",
		    ivmob->cmd_origin == IVMOB_CMD_ORIGIN_REMOTE ?
		    "remote" : "local");
	}
	if (ivmob->cmd_origin == IVMOB_CMD_ORIGIN_REMOTE)
		status->rs_mode = IVMOB_MANUAL;
	else
		switch (status->operational_mode) {
		case IVMOB_MODE_INACTIVE:
			status->rs_mode = IVMOB_MOTORS_OFF;
			break;
		case IVMOB_MODE_ACTIVE:
			break;
		case IVMOB_MODE_SAFETY:
			status->rs_mode = IVMOB_EMERGENCY;
			if (prevMode != IVMOB_MODE_SAFETY)
				printf("** safety error: %s\n",
				  ivmob_safety_code(ivmob->safety_error_code));
			break;
		case IVMOB_MODE_WAIT_RESET:
			status->rs_mode = IVMOB_MOTORS_OFF;
			break;
		}
	status->safety_code = ivmob->safety_error_code;
	status->battery_level = ivmob->battery_level*100.0/8.0;
	status->v = ivmob->v/1000.0;
	status->w = ivmob->w/1000.0;
	status->temperature = ivmob->temperature;
	status->front_left_wheel_speed =
	    (double)ivmob->front_left_wheel_speed / 1e3;
	status->front_right_wheel_speed =
	    (double)ivmob->front_right_wheel_speed / 1e3;
	status->rear_left_wheel_speed =
	    (double)ivmob->rear_left_wheel_speed / 1e3;
	status->rear_right_wheel_speed =
	    (double)ivmob->rear_right_wheel_speed / 1e3;
	status->front_left_current =
	    (double)ivmob->front_left_current / 1e3;
	status->front_right_current =
	    (double)ivmob->front_right_current / 1e3;
	status->rear_left_current =
	    (double)ivmob->rear_left_current / 1e3;
	status->rear_right_current =
	    (double)ivmob->rear_right_current / 1e3;
	status->mmi_state = ivmob->mmi;

	prevMode = ivmob->operational_mode;
	prevOrigin = ivmob->cmd_origin;
	ivmob_unlock(ivmob);
}

/*------------------------------------------------------------------------
 *
 * odo  --  1st codel of permanent activity (called BEFORE other activities)
 *
 * Description: Reads all robots state and computes odometry
 *
 * Reports:      OK
 *              S_ivmob_EMERGENCY_STOP
 *
 * Returns:    OK or ERROR
 */

STATUS
odo(int *report)
{
	GENPOS_CART_REF *refId = &SDI_F->ref;
	GENPOS_CART_STATE *robot = &SDI_F->robot;
	IVMOB_GYRO *gyro = &SDI_F->gyro;
	double direction;
#ifdef GYRO
	uint32_t date;
#endif
#if DEBUG>=1
	static int count = 0;
#endif

	if (ivmob == NULL)
		return OK;		/* not initialized yet */

	/* Read data */
	ivmobDataUpdate(&SDI_F->status);
	if (ivmobVelocityGet() != OK)
		return ERROR;
	if (refId->backFlag == GENPOS_FORWARD_MOTION)
		direction = 1;
	else if (refId->backFlag == GENPOS_BACKWARD_MOTION)
		direction = -1;
	else
		direction = 0;
	robot->xRef = robot->xRob
		+ direction*IVMOB_DIST_TO_CONTROLLED_POINT*cos(robot->theta);
	robot->yRef = robot->yRob
		+ direction*IVMOB_DIST_TO_CONTROLLED_POINT*sin(robot->theta);

	odoProba(&SDI_F->robot, &SDI_F->var,
	    SDI_F->kinematics.axisWidth, SDI_F->varParams.coeffLinAng,
		   EXEC_TASK_PERIOD(IVMOB_MOTIONTASK_NUM));

	/* Gyro */
#ifdef GYRO
	if (gyroId != NULL /* && gyro->currentMode != IVMOB_GYRO_OFF */) {
		if (gyroRead(gyroId, &gyro->gyroTheta,
			     &gyro->gyroOmega, &date) != 0) {
			gyro->currentMode = IVMOB_GYRO_OFF;
			gyroEnd(gyroId);
			gyroId = NULL;
		} else {
			/* XXX gyro is mounted upside down */
			gyro->gyroTheta = - gyro->gyroTheta;
			gyro->gyroOmega = - gyro->gyroOmega;
		}
	}
#endif
	if (gyro->currentMode == IVMOB_GYRO_OFF ||
	    (gyro->currentMode == IVMOB_GYRO_ON_IF_MOTION
		&& fabs(robot->w) < 0.017
		&& fabs(robot->v) < 0.01)) {
		/* -- odo is reference */

		/* reset gyro offset to match odo */
		gyro->gyroToRobotOffset = angleLimit(robot->theta
		    - gyro->gyroTheta);
		/* switch back to odo */
		if (gyro->gyroOn) {
#if DEBUG>=1
			printf ("-> odo (%5.2f d) v %5.2f w%5.3f\n",
			    RAD_TO_DEG(robot->theta), robot->v, robot->w);
#endif
			gyro->gyroOn = FALSE;
		}
	} else {
		/* -- gyro is reference */
		robot->theta = angleLimit(gyro->gyroTheta +
		    gyro->gyroToRobotOffset);
		/* switch back to gyro */
		if (!gyro->gyroOn) {
#if DEBUG>=1
			printf ("-> gyro (%5.2f d) v %5.2f w%5.3f\n",
			    RAD_TO_DEG(robot->theta), robot->v, robot->w);
#endif
			gyro->gyroOn = TRUE;
			ivmobDataStrId->gyroAsserv.first = 1;
		}
	}

	/*debug */
#if DEBUG>=1
	if (count==10) {
		printf ("MODE %s. gyro %5.3f + offset %5.3f = %5.3f =? %5.3f\n",
		    gyro->gyroOn ? "GYRO" : "ODO",
		    RAD_TO_DEG(gyro->gyroTheta),
		    RAD_TO_DEG(gyro->gyroToRobotOffset),
		    RAD_TO_DEG(angleLimit(gyro->gyroTheta +
			    gyro->gyroToRobotOffset)),
		    RAD_TO_DEG(robot->theta));
		count=0;
	}
	count++;
#endif
	if (SDI_F->odometryMode == IVMOB_ODO_3D) {
	  ivmobOdo3d (report, EXEC_TASK_PERIOD(IVMOB_MOTIONTASK_NUM));


	}
	return OK;
}

/*----------------------------------------------------------------------*/

/**
 ** Actual tracking code
 **/
static void
track(int *report, double *vRef, double *wRef)
{
	GENPOS_CART_REF *refId = &SDI_F->ref;
	GENPOS_CART_STATE *robot = &SDI_F->robot;
	GENPOS_TRACK *trackId = &SDI_F->trackStr;
	static double prevV, prevW;
	static BOOL prevDrifted = FALSE;
	BOOL drifted = FALSE;

	switch (trackId->trackMode) {
	case GENPOS_TRACK_POS:
		if (refId->dataType != GENPOS_POS_DATA) {
			fprintf(stderr, "wrong data type for TRACK_POS\n");
			*report = S_ivmob_BAD_REF;
			return;
		}
		drifted = contLawCartPositionControl(&ivmobDataStrId->cmd,
		    refId, robot, EXEC_TASK_PERIOD(IVMOB_MOTIONTASK_NUM),
		    prevV, prevW, vRef, wRef);
		break;

	case GENPOS_TRACK_CONFIG:
		if (refId->dataType != GENPOS_POS_AND_SPEED_DATA) {
			fprintf(stderr, "wrong data type for TRACK_CONFIG\n");
			*report = S_ivmob_BAD_REF;
			return;
		}
		drifted = contLawCartConfigControl(&ivmobDataStrId->cmd,
		    refId, robot, EXEC_TASK_PERIOD(IVMOB_MOTIONTASK_NUM),
		    prevV, prevW, vRef, wRef);
		break;

	case GENPOS_TRACK_SPEED:
		if (refId->dataType != GENPOS_SPEED_DATA) {
			fprintf(stderr, "wrong data type for TRACK_SPEED %d\n",
			    refId->dataType);
			*report = S_ivmob_BAD_REF;
			return;
		}
		*vRef = (refId->backFlag == GENPOS_FORWARD_MOTION ? refId->v :
		    -refId->v);
		*wRef = refId->w;
		break;
	default:
		*report = S_ivmob_BAD_REF;
		break;
	}
	if (drifted && !prevDrifted) {
		printf("odoAndAsserv: important drift!\n");
		printf("ref: %.2lf %.2lf %.2lf\n",
		       refId->x, refId->y, refId->theta);
		printf("pos: %.2lf %.2lf %.2lf\n",
		       robot->xRef, robot->yRef, robot->theta);
	}
	if (!drifted && prevDrifted)
		printf("odoAndAsserv: drift cancelled\n");
	prevDrifted = drifted;

	prevV = *vRef;
	prevW = *wRef;
}

/*----------------------------------------------------------------------*/

/*
 * Filter velocities to keep accelerations within max bounds
 *
 * don't use real time but rather module's period, so that if some periods
 * are lost it doesn't accelerate too much afterwards
 *
 * When moving along an arc (keep_radius = true) adjust angular speed
 * to respect the circle's radius
 */
void
bound_accels(double t, double *vel_reference, double *ang_reference)
{
	IVMOB_MAX_ACCEL *acc = &ivmobDataStrId->maxAccel;
	const double epsilon = 1e-3;
	int keep_radius = (fabs(*ang_reference) > epsilon &&
	    fabs(*vel_reference) > epsilon);
	double radius;

	if (keep_radius)
		radius = *vel_reference / *ang_reference;

	// linear velocity
	double sign_accel = *vel_reference > acc->prev_vel_command ?
	    +1.0 : -1.0;
	if (acc->prev_vel_command * *vel_reference < 0)
		*vel_reference = 0.0;

	double max_accel = IVMOB_DEFAULT_MAXIMUM_ACCEL;
	double max_vel = acc->prev_vel_command +
	    sign_accel * max_accel * IVMOB_SEC_PERIOD;
	*vel_reference = sign_accel *
	    fmin(sign_accel * *vel_reference, sign_accel * max_vel);
	acc->prev_vel_command = *vel_reference;

	// angular velocity to preserve radius
	if (keep_radius) *ang_reference = *vel_reference / radius;
}

/*----------------------------------------------------------------------*/

/*
 * Control robot rotations
 *
 * The internal control laws are so inefficient that the robot
 * is not able to follow an arc of circle of a given radius
 *
 * So implement a control using an external (gyroscop) measure of the
 * yaw rate to adjust the speeds to make that better
 *
 * inputs:
 *   t - current time
 *   vel_reference  - desired linear velocity
 *   yawr_reference - desired angular velocity
 *   yawr_measure   - mesured angular velocity
 *   yaw_measure    - mesured angular position
 * outputs
 *   report - Genom error status if any
 *   yawr_command - corrected angular velocity to be sent to the robot
 *
 * also access the gyroAsserv member of the IDS directly.
 */

void
control_yaw(int *report, double t, double vel_reference, double yawr_reference,
	    double yawr_measure, double yaw_measure, double *yawr_command)
{
	IVMOB_GYRO_ASSERV *gyro = &ivmobDataStrId->gyroAsserv;
	*yawr_command = gyro->prev_command;

	if (gyro->first)
	{
		*yawr_command = yawr_reference;
		gyro->first = 0;
		gyro->enabled = 1;
		gyro->jump_t = 0.;
	} else
	{
		double dt = t - gyro->prev_t;

		if (yawr_reference == 0.0)
		{
			if (vel_reference != 0.0)
			{ // straight -> directly asserv on angle
				if (gyro->straight)
				{
					if (yawr_measure == gyro->prev_measure) return;

					// PI controller on angle commanded in speed
					// I is needed to remove static error
					// D is not needed because only used to go straight (small errors)
					double error_s = (gyro->straight_angle-yaw_measure);
					if (error_s > PI) error_s -= 2*PI;
					if (error_s < -PI) error_s += 2*PI;
					gyro->integral_s += error_s * (t-gyro->prev_t);
					*yawr_command =
						IVMOB_KP_GYRO_THETA*dt * error_s +
						IVMOB_KI_GYRO_THETA*dt * gyro->integral_s;
				} else
				{
					*yawr_command = 0.0;
					gyro->straight = 1;
					gyro->straight_angle = yaw_measure;
					gyro->integral_s = 0.0;
				}
			} else
			{ // do nothing, command (0,0) is correctly respected
				*yawr_command = 0.0;
				gyro->straight = 0;
			}
		} else
		{
			const double ref_ratio = 1.5;
			const double noise_th = 0.1;

			// if changing sign or increasing from around 0
			// then directly apply a priori command
			if ((yawr_reference*gyro->prev_reference < 0) ||
			    ((fabs(gyro->prev_reference) < noise_th) && (fabs(yawr_reference) >= noise_th*ref_ratio)))
			{
				*yawr_command = yawr_reference;
				gyro->jump_t = t;
			}
			// if changing more than 50%
			// then directly apply extrapolated command
			if ((fabs(gyro->prev_reference) >= noise_th) &&
			    (yawr_reference*gyro->prev_reference > 0) &&
			    ((yawr_reference/gyro->prev_reference > ref_ratio) || (gyro->prev_reference/yawr_reference > ref_ratio)))
			{
				*yawr_command = gyro->prev_command * yawr_reference / gyro->prev_reference;
				gyro->jump_t = t;
			}
			// otherwise if delay from last jump has elapsed
			// P controller on speed commanded in speed
			if (t-gyro->jump_t >= IVMOB_DELAY_GYRO_OMEGA)
			{
				if (yawr_measure == gyro->prev_measure) return;

				*yawr_command =
					gyro->prev_command +
					IVMOB_KP_GYRO_OMEGA*dt * (yawr_reference-yawr_measure);
			} else
			// if we are in the lock period and we didn't just reenter it,
			// then we have to follow the reference changes
			if (t != gyro->jump_t)
			{
				if (fabs(gyro->prev_reference) < noise_th || fabs(yawr_reference - gyro->prev_reference) < 1e-10)
					*yawr_command = gyro->prev_command;
				else
					*yawr_command = gyro->prev_command * yawr_reference / gyro->prev_reference;
			}
			gyro->straight = 0;
		}
	}

//*yawr_command = yawr_reference; // TEST remove asserv

	gyro->prev_reference = yawr_reference;
	gyro->prev_measure = yawr_measure;
	gyro->prev_command = *yawr_command;
	gyro->prev_t = t;
}


/*------------------------------------------------------------------------
 *
 * asserv  --  2nd codel of permanent activity (called AFTER other activities)
 *
 * Description: Computes reference speeds and apply them
 *
 * Reports:      OK
 *              S_ivmob_EMERGENCY_STOP
 *
 * Returns:    OK or ERROR
 */

STATUS
asserv(int *report)
{
	GENPOS_CART_REF *refId = &SDI_F->ref;
	GENPOS_CART_STATE *robot = &SDI_F->robot;
	double vRef, wRef;

	if (ivmob == NULL)
		return OK;		/* not initialized yet */

	*report = OK;
	switch (SDI_F->status.rs_mode) {
	case IVMOB_MOTORS_OFF:
	case IVMOB_EMERGENCY:
	case IVMOB_MANUAL:
		/* No motion possible */
		return OK;

	case IVMOB_IDLE:
		vRef = 0.0;
		wRef = 0.0;
		refId->x = robot->xRef;
		refId->y = robot->yRef;
		refId->theta = robot->theta;
		break;

	case IVMOB_TRACK:
		track(report, &vRef, &wRef);
		break;

	default:
		*report = S_ivmob_BAD_REF;
		return ERROR;
	}
	if (*report != OK) {
		/* In case an error occured,
		   stop the robot and the tracking */
		SDI_F->status.rs_mode = IVMOB_IDLE;
		SDI_F->trackStr.trackMode = GENPOS_NO_TRACKING;
		vRef = 0;
		wRef = 0;
	}

	SDI_F->vReference = vRef;
	SDI_F->wReference = wRef;

	H2TIMESPEC tv;
	h2GetTimeSpec(&tv);
	double t = tv.tv_sec + tv.tv_nsec*1e-9;
	//if (SDI_F->status.rs_mode == IVMOB_TRACK)
	bound_accels(t, &vRef, &wRef);

#ifdef GYRO
	if (SDI_F->gyro.gyroOn)
		control_yaw(report, t, vRef, wRef, SDI_F->gyro.gyroOmega, SDI_F->gyro.gyroTheta, &wRef);
	//(SDI_F->status.rs_data[0].yaw_rate + SDI_F->status.rs_data[1].yaw_rate)/2.0; // do not use internal gyros, they are wrong
#endif
	if (ivmobVelocitySet(vRef, wRef, report) != OK)
		return ERROR;
	updatePomPosters();

	return OK;
}

/*------------------------------------------------------------------------
 * Init
 *
 * Description:
 *
 * Reports:      OK
 */

/* ivmobInitMain  -  codel EXEC of Init
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
ivmobInitStart(GEN_STRING32 *device, int *report)
{

	*report = OK;
	/* Disconnect first if already connected */
	if (ivmob != NULL) {
		ivmobCmdInactive(ivmob, 0);
		ivmobEnd(ivmob);
	}

	ivmob = ivmobInit(device->name, true);

	if (ivmob == NULL) {
		*report = S_ivmob_IVMOBLIB_ERROR;
		return ETHER;
	}
	ivmob->operational_mode = -1;
	if (ivmobRqstPeriodicMessages(ivmob, 1, 0, 0, 1, 1, 1, -1) != 0) {
		*report = S_ivmob_IVMOBLIB_ERROR;
		return ETHER;
	}
	/* read kinematics from robot */
	ivmobRqstWheelSize(ivmob, -1);
	ivmobRqstWheelTrackWidth(ivmob, -1);
	return EXEC;
}

ACTIVITY_EVENT
ivmobInitMain(GEN_STRING32 *device, int *report)
{
	IVMOB_KINEMATICS_STR *kinematics = &SDI_F->kinematics;
	IVMOB_DYNAMIC_STR *dynamics = &SDI_F->dynamics;


	/* Wait for first update from robot */
	if (ivmob->operational_mode < 0)
		return EXEC;

	/* dynamics */
	dynamics->vMax = IVMOB_DEFAULT_MAXIMUM_VELOCITY;
	dynamics->wMax = IVMOB_DEFAULT_MAXIMUM_YAW_RATE;
	dynamics->linAccelMax = IVMOB_DEFAULT_MAXIMUM_ACCEL;
	dynamics->angAccelMax = IVMOB_DEFAULT_MAXIMUM_YAW_ACCEL;

	/* kinematics */
	kinematics->wheelRadius = ivmob->wheel_diameter / 2000.0;
	kinematics->axisWidth = ivmob->wheel_track_width / 1000.0;
	return ETHER;
}

/*------------------------------------------------------------------------
 * Activate
 *
 * Description:
 *
 * Reports:      OK
 */
ACTIVITY_EVENT
ivmobActivateStart(int *active, int *report)
{
	int ret;

	*report = OK;
	if (*active)
		ret = ivmobCmdActive(ivmob, 0);
	else
		ret = ivmobCmdInactive(ivmob, 0);

	if (ret < 0) {
		*report = S_ivmob_IVMOBLIB_ERROR;
		return END;
	}
	return EXEC;
}

/* ivmobActivateMain  -  codel EXEC of Init
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
ivmobActivateMain(int *active, int *report)
{
	IVMOB_STATUS_STR *status = &SDI_F->status;

	/* Check motors status */
	if (ivmob->operational_mode == IVMOB_MODE_SAFETY) {
		*report = S_ivmob_EMERGENCY_STOP;
		    return ETHER;
	}
	if (*active) {
		if (ivmob->operational_mode != IVMOB_MODE_ACTIVE)
			return EXEC;
	} else {
		if (ivmob->operational_mode != IVMOB_MODE_INACTIVE)
			return EXEC;
	}
	printf("Motors ON\n");
	status->rs_mode = IVMOB_IDLE;
	return ETHER;
}

/*------------------------------------------------------------------------
 * SetPom
 *
 * Description:
 *
 * Reports:      OK
 *              S_ivmob_POSTER_NOT_FOUND
 */

/* setPom  -  codel EXEC of SetPom
   Returns:  EXEC END ETHER FAIL ZOMBIE */

ACTIVITY_EVENT
setPom(GENPOS_POSTER_NAME *posterName, int *report)
{
	POSTER_ID id;

	if (posterFind(posterName->name, &id) != OK) {
		logMsg("posterFind(%s) failed\n", posterName->name);
		*report = S_ivmob_POSTER_NOT_FOUND;
	} else
		pomPosterId = id;
	return ETHER;
}

/*----------------------------------------------------------------------*/
/*
 * Compute speed
 *
 * sets robot.v and robot.w in the SDI from motion in the motor structs.
 */


static STATUS
ivmobVelocityGet(void)
{
	GENPOS_CART_STATE *robot = &SDI_F->robot;
	double leftv, rightv, r, l;

	r = SDI_F->kinematics.wheelRadius;
	l = SDI_F->kinematics.axisWidth;

	/* Mean angular speed of front and back wheels */
	leftv = r*(ivmob->front_left_wheel_speed +
	    ivmob->rear_left_wheel_speed) / 2000.0;
	rightv = r*(ivmob->front_right_wheel_speed +
	    ivmob->rear_right_wheel_speed) / 2000.0;

	/* Angular speed of the robot */
	robot->w = -(leftv - rightv) / l;
	/* linear speed */
	robot->v = (leftv + rightv) / 2.0;

#if 0
	if (abs(ivmob->v) > 10 || abs(ivmob->w) > 10) {
		printf("wheels %d %d %d %d %d %.3lf %.3lf\n",
		    ivmob->front_left_wheel_speed,
		    ivmob->rear_left_wheel_speed,
		    ivmob->front_right_wheel_speed,
		    ivmob->rear_right_wheel_speed,
		    leftv, rightv);
		printf("ivmob received  %.3lf %.3lf / computed %.3lf %.3lf\n",
		    ivmob->v/1000.0, ivmob->w/1000.0, robot->v, robot->w);
	}
#endif
	return OK;
}

/*----------------------------------------------------------------------*/
/*
 * Send given linear and angular speeds to the hardware
 *
 */
static STATUS
ivmobVelocitySet(double v, double w, int *report)
{
	struct IVMOB_DYNAMIC_STR *dynamics = &SDI_F->dynamics;
	uint16_t vmm, wmrad;

	/* Last safety check, in case uninitialized values are used */
	if (fabs(v) > dynamics->vMax) {
		fprintf(stderr, "WARNING: fabs(v): > VMAX: %lf\n", v);
		v = v > 0.0 ? dynamics->vMax : -dynamics->vMax;
	}
	/* Save the set velocities */
	ivmobDataStrId->vCommand = v;
	/* milimeter/s */
	vmm = (int)(v*1000);

	if (fabs(w) > dynamics->wMax) {
		fprintf(stderr, "WARNING: w > WMAX: %lf\n", w);
		w = w > 0.0 ? dynamics->wMax : -dynamics->wMax;
	}

	/* Save the set velocities */
	ivmobDataStrId->wCommand = w;
	/* mRad/s */
	wmrad = (int)(w*1000);

#if 0
	if (abs(vmm) > 10 || abs(wmrad) > 10)
		printf("\n velocitySet %.03lf %.03lf %d %d\n",
		    v, w, vmm, wmrad);
#endif
	/* Send commands to motors if ON and we control the robot*/
	if (ivmob->cmd_origin == IVMOB_CMD_ORIGIN_LOCAL &&
	    ivmob->operational_mode == IVMOB_MODE_ACTIVE)
		/* don't check the return value, since some messges get lost */
		ivmobCmdSpeed(ivmob, vmm, wmrad, 50);

	return OK;
}

/*----------------------------------------------------------------------*/
/*
 * Initialize POM
 */
static STATUS
initPomPosters(void)
{
	POM_ME_POS *mePos = &ivmobDataStrId->mePos;

	mePos->kind = POM_ME_ABSOLUTE;
	mePos->date1 = -1;
	mePos->confidence = 0.2;	// XXX used?
	mePos->main.var.voo     = IVMOB_VAR_TT;
	mePos->main.var.vpp     = IVMOB_VAR_TT;
	mePos->main.var.vrr     = IVMOB_VAR_TT;
	mePos->main.var.vxx     = IVMOB_VAR_XX;
	mePos->main.var.vyy     = IVMOB_VAR_YY;
	mePos->main.var.vzz     = IVMOB_VAR_XX;

	return OK;
}

/*----------------------------------------------------------------------*/

/*
 * Update POM poster
 */
static STATUS
updatePomPosters(void)
{
	POM_ME_POS *mePos = &SDI_F->mePos;
	GENPOS_CART_STATE *robot;
	IVMOB_CART_3DSTATE *config3dId;
	POM_POS pomPos;

	if (pomPosterId != NULL) {
		if(ivmobPOM_POSPosterRead(pomPosterId, &pomPos) == OK) {
			mePos->date1 = pomPos.date;
		} else {
			logMsg("ivmobPOM_POSPosterRead() failed\n");
			pomPosterId = NULL;
			mePos->date1 = -1;
		}
	} else
		mePos->date1 = -1;

	robot = &ivmobDataStrId->robot;
	mePos->main.euler.yaw = robot->theta;
	mePos->main.euler.x   = robot->xRob;
	mePos->main.euler.y   = robot->yRob;

	config3dId = &ivmobDataStrId->robot3d;
	mePos->main.euler.pitch = config3dId->pitch;
	mePos->main.euler.roll = config3dId->roll;
	mePos->main.euler.z = config3dId->zRob;

	/* speeds are mostly zero, just update the linear
	   and angular components */
	mePos->VLocal.euler.x   = robot->v;
	mePos->VLocal.euler.yaw = robot->w;

	/* all variances and co-variances remain zero */

	return OK;
}

/*------------------------------------------------------------------------
 * Gyro
 *
 * Description: Setup gyroscope
 *
 * Reports:      OK
 *              S_ivmob_GYRO_ERROR
 */

/* ivmobGyroExec  -  codel EXEC of Gyro
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
ivmobGyroExec(IVMOB_GYRO_PARAMS *gyroParams, int *report)
{
	IVMOB_GYRO *gyro = &SDI_F->gyro;

#ifdef GYRO
	if (gyroId == NULL) {
		gyroId = gyroInit(gyroParams->type, gyroParams->port,
			    gyroParams->latitude, gyroParams->woffset);
		if (gyroId == NULL) {
			*report = S_ivmob_GYRO_ERROR;
			gyro->params.mode = IVMOB_GYRO_OFF;
			gyro->currentMode = IVMOB_GYRO_OFF;
			return ETHER;
		}
	}

	/* read gyro once */
	if (gyroReadAngle(gyroId, &gyro->gyroTheta) != 0) {
		*report = S_ivmob_GYRO_ERROR;
		gyro->params.mode = IVMOB_GYRO_OFF;
		gyro->currentMode = IVMOB_GYRO_OFF;
		gyroEnd(gyroId);
		gyroId = NULL;
		return ETHER;
	} else {
		/* XXX gyro is mounted upside down */
		gyro->gyroTheta = - gyro->gyroTheta;
		gyro->gyroOmega = - gyro->gyroOmega;
	}

	/* reset gyro offset to match odo */
	gyro->gyroToRobotOffset = SDI_F->robot.theta - gyro->gyroTheta;

	gyro->currentMode = gyro->params.mode;

#else
	printf("ivmobGyroExec: ivmob module compiled without gyro support. "
	    "Use -DGYRO\n");
	gyro->params.mode = IVMOB_GYRO_OFF;
	gyro->currentMode = IVMOB_GYRO_OFF;
	*report = S_ivmob_GYRO_ERROR;
#endif
	return ETHER;
}
